@@ -23,6 +23,8 @@ import signal
 import sys
 import time
 
+import psutil
+
 from mock import Mock
 
 from twisted.internet import defer
@@ -33,8 +35,8 @@ from twisted.python import runtime
 from twisted.python import util
 from twisted.trial import unittest
 
-from buildbot_worker import util as bsutil
 from buildbot_worker import runprocess
+from buildbot_worker import util as bsutil
 from buildbot_worker.exceptions import AbandonChain
 from buildbot_worker.test.fake.workerforbuilder import FakeWorkerForBuilder
 from buildbot_worker.test.util import compat
@@ -47,15 +49,15 @@ def catCommand():
 
 
 def stdoutCommand(output):
-    return [sys.executable, '-c', 'import sys; sys.stdout.write("%s\\n")' % output]
+    return [sys.executable, '-c', 'import sys; sys.stdout.write("{0}\\n")'.format(output)]
 
 
 def stderrCommand(output):
-    return [sys.executable, '-c', 'import sys; sys.stderr.write("%s\\n")' % output]
+    return [sys.executable, '-c', 'import sys; sys.stderr.write("{0}\\n")'.format(output)]
 
 
 def sleepCommand(dur):
-    return [sys.executable, '-c', 'import time; time.sleep(%d)' % dur]
+    return [sys.executable, '-c', 'import time; time.sleep({0})'.format(dur)]
 
 
 def scriptCommand(function, *args):
@@ -105,96 +107,117 @@ class TestRunProcess(BasedirMixin, unittest.TestCase):
         self.assertIsInstance(s.command[0], bytes)
         self.assertIsInstance(s.fake_command[0], bytes)
 
+    @defer.inlineCallbacks
     def testStart(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(b, stdoutCommand('hello'), self.basedir)
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            self.assertTrue({'stdout': nl('hello\n')} in b.updates, b.show())
-            self.assertTrue({'rc': 0} in b.updates, b.show())
-        d.addCallback(check)
-        return d
+        self.assertTrue({'stdout': nl('hello\n')} in b.updates, b.show())
+        self.assertTrue({'rc': 0} in b.updates, b.show())
 
+    @defer.inlineCallbacks
     def testNoStdout(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(
             b, stdoutCommand('hello'), self.basedir, sendStdout=False)
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            self.failIf({'stdout': nl('hello\n')} in b.updates, b.show())
-            self.assertTrue({'rc': 0} in b.updates, b.show())
-        d.addCallback(check)
-        return d
+        self.failIf({'stdout': nl('hello\n')} in b.updates, b.show())
+        self.assertTrue({'rc': 0} in b.updates, b.show())
 
+    @defer.inlineCallbacks
     def testKeepStdout(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(
             b, stdoutCommand('hello'), self.basedir, keepStdout=True)
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            self.assertTrue({'stdout': nl('hello\n')} in b.updates, b.show())
-            self.assertTrue({'rc': 0} in b.updates, b.show())
-            self.assertEqual(s.stdout, nl('hello\n'))
-        d.addCallback(check)
-        return d
+        self.assertTrue({'stdout': nl('hello\n')} in b.updates, b.show())
+        self.assertTrue({'rc': 0} in b.updates, b.show())
+        self.assertEqual(s.stdout, nl('hello\n'))
 
+    @defer.inlineCallbacks
     def testStderr(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(b, stderrCommand("hello"), self.basedir)
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            self.failIf({'stderr': nl('hello\n')} not in b.updates, b.show())
-            self.assertTrue({'rc': 0} in b.updates, b.show())
-        d.addCallback(check)
-        return d
+        self.failIf({'stderr': nl('hello\n')} not in b.updates, b.show())
+        self.assertTrue({'rc': 0} in b.updates, b.show())
 
+    @defer.inlineCallbacks
     def testNoStderr(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(
             b, stderrCommand("hello"), self.basedir, sendStderr=False)
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            self.failIf({'stderr': nl('hello\n')} in b.updates, b.show())
-            self.assertTrue({'rc': 0} in b.updates, b.show())
-        d.addCallback(check)
-        return d
+        self.failIf({'stderr': nl('hello\n')} in b.updates, b.show())
+        self.assertTrue({'rc': 0} in b.updates, b.show())
 
+    @defer.inlineCallbacks
+    def test_incrementalDecoder(self):
+        b = FakeWorkerForBuilder(self.basedir)
+        b.unicode_encoding = "utf-8"
+        s = runprocess.RunProcess(
+            b, stderrCommand("hello"), self.basedir, sendStderr=True)
+        pp = runprocess.RunProcessPP(s)
+        # u"\N{SNOWMAN} when encoded to utf-8 bytes is b"\xe2\x98\x83"
+        pp.outReceived(b"\xe2")
+        pp.outReceived(b"\x98\x83")
+        pp.errReceived(b"\xe2")
+        pp.errReceived(b"\x98\x83")
+        yield s.start()
+
+        self.assertTrue({'stderr': u"\N{SNOWMAN}"} in b.updates)
+        self.assertTrue({'stdout': u"\N{SNOWMAN}"} in b.updates)
+        self.assertTrue({'rc': 0} in b.updates, b.show())
+
+    @defer.inlineCallbacks
+    def testInvalidUTF8(self):
+        b = FakeWorkerForBuilder(self.basedir)
+        b.unicode_encoding = "utf-8"
+        s = runprocess.RunProcess(
+            b, stderrCommand("hello"), self.basedir, sendStderr=True)
+        pp = runprocess.RunProcessPP(s)
+        INVALID_UTF8 = b"\xff"
+        with self.assertRaises(UnicodeDecodeError):
+            INVALID_UTF8.decode('utf-8')
+        pp.outReceived(INVALID_UTF8)
+        yield s.start()
+        stdout = [up['stdout'] for up in b.updates if 'stdout' in up][0]
+        # On Python < 2.7 bytes is used, on Python >= 2.7 unicode
+        self.assertIn(stdout, (b'\xef\xbf\xbd', u'\ufffd'))
+        self.assertTrue({'rc': 0} in b.updates, b.show())
+
+    @defer.inlineCallbacks
     def testKeepStderr(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(
             b, stderrCommand("hello"), self.basedir, keepStderr=True)
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            self.assertTrue({'stderr': nl('hello\n')} in b.updates, b.show())
-            self.assertTrue({'rc': 0} in b.updates, b.show())
-            self.assertEqual(s.stderr, nl('hello\n'))
-        d.addCallback(check)
-        return d
+        self.assertTrue({'stderr': nl('hello\n')} in b.updates, b.show())
+        self.assertTrue({'rc': 0} in b.updates, b.show())
+        self.assertEqual(s.stderr, nl('hello\n'))
 
+    @defer.inlineCallbacks
     def testStringCommand(self):
         b = FakeWorkerForBuilder(self.basedir)
         # careful!  This command must execute the same on windows and UNIX
         s = runprocess.RunProcess(b, 'echo hello', self.basedir)
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            self.assertTrue({'stdout': nl('hello\n')} in b.updates, b.show())
-            self.assertTrue({'rc': 0} in b.updates, b.show())
-        d.addCallback(check)
-        return d
+        self.assertTrue({'stdout': nl('hello\n')} in b.updates, b.show())
+        self.assertTrue({'rc': 0} in b.updates, b.show())
 
     def testObfuscatedCommand(self):
         b = FakeWorkerForBuilder(self.basedir)
@@ -204,6 +227,7 @@ class TestRunProcess(BasedirMixin, unittest.TestCase):
         self.assertEqual(s.command, [b'abcd'])
         self.assertEqual(s.fake_command, [b'ABCD'])
 
+    @defer.inlineCallbacks
     def testMultiWordStringCommand(self):
         b = FakeWorkerForBuilder(self.basedir)
         # careful!  This command must execute the same on windows and UNIX
@@ -212,27 +236,23 @@ class TestRunProcess(BasedirMixin, unittest.TestCase):
 
         # no quoting occurs
         exp = nl('Happy Days and Jubilation\n')
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            self.assertTrue({'stdout': exp} in b.updates, b.show())
-            self.assertTrue({'rc': 0} in b.updates, b.show())
-        d.addCallback(check)
-        return d
+        self.assertTrue({'stdout': exp} in b.updates, b.show())
+        self.assertTrue({'rc': 0} in b.updates, b.show())
 
+    @defer.inlineCallbacks
     def testInitialStdinUnicode(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(
             b, catCommand(), self.basedir, initialStdin=u'hello')
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            self.assertTrue({'stdout': nl('hello')} in b.updates, b.show())
-            self.assertTrue({'rc': 0} in b.updates, b.show())
-        d.addCallback(check)
-        return d
+        self.assertTrue({'stdout': nl('hello')} in b.updates, b.show())
+        self.assertTrue({'rc': 0} in b.updates, b.show())
 
+    @defer.inlineCallbacks
     def testMultiWordStringCommandQuotes(self):
         b = FakeWorkerForBuilder(self.basedir)
         # careful!  This command must execute the same on windows and UNIX
@@ -245,14 +265,12 @@ class TestRunProcess(BasedirMixin, unittest.TestCase):
             exp = nl('"Happy Days and Jubilation"\n')
         else:
             exp = nl('Happy Days and Jubilation\n')
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            self.assertTrue({'stdout': exp} in b.updates, b.show())
-            self.assertTrue({'rc': 0} in b.updates, b.show())
-        d.addCallback(check)
-        return d
+        self.assertTrue({'stdout': exp} in b.updates, b.show())
+        self.assertTrue({'rc': 0} in b.updates, b.show())
 
+    @defer.inlineCallbacks
     def testTrickyArguments(self):
         # make sure non-trivial arguments are passed verbatim
         b = FakeWorkerForBuilder(self.basedir)
@@ -267,14 +285,12 @@ class TestRunProcess(BasedirMixin, unittest.TestCase):
         ]
 
         s = runprocess.RunProcess(b, printArgsCommand() + args, self.basedir)
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            self.assertTrue({'stdout': nl(repr(args))} in b.updates, b.show())
-            self.assertTrue({'rc': 0} in b.updates, b.show())
-        d.addCallback(check)
-        return d
+        self.assertTrue({'stdout': nl(repr(args))} in b.updates, b.show())
+        self.assertTrue({'rc': 0} in b.updates, b.show())
 
+    @defer.inlineCallbacks
     @compat.skipUnlessPlatformIs("win32")
     def testPipeString(self):
         b = FakeWorkerForBuilder(self.basedir)
@@ -283,45 +299,43 @@ class TestRunProcess(BasedirMixin, unittest.TestCase):
             ' -c "import sys; sys.stdout.write(\'b\\na\\n\')" | sort'
         s = runprocess.RunProcess(b, cmd, self.basedir)
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            self.assertTrue({'stdout': nl('a\nb\n')} in b.updates, b.show())
-            self.assertTrue({'rc': 0} in b.updates, b.show())
-        d.addCallback(check)
-        return d
+        self.assertTrue({'stdout': nl('a\nb\n')} in b.updates, b.show())
+        self.assertTrue({'rc': 0} in b.updates, b.show())
 
+    @defer.inlineCallbacks
     def testCommandTimeout(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(b, sleepCommand(10), self.basedir, timeout=5)
         clock = task.Clock()
         s._reactor = clock
+
         d = s.start()
-
-        def check(ign):
-            self.assertTrue(
-                {'stdout': nl('hello\n')} not in b.updates, b.show())
-            self.assertTrue({'rc': FATAL_RC} in b.updates, b.show())
-        d.addCallback(check)
         clock.advance(6)
-        return d
+        yield d
 
+        self.assertTrue(
+            {'stdout': nl('hello\n')} not in b.updates, b.show())
+        self.assertTrue({'rc': FATAL_RC} in b.updates, b.show())
+
+    @defer.inlineCallbacks
     def testCommandMaxTime(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(b, sleepCommand(10), self.basedir, maxTime=5)
         clock = task.Clock()
         s._reactor = clock
-        d = s.start()
 
-        def check(ign):
-            self.assertTrue(
-                {'stdout': nl('hello\n')} not in b.updates, b.show())
-            self.assertTrue({'rc': FATAL_RC} in b.updates, b.show())
-        d.addCallback(check)
+        d = s.start()
         clock.advance(6)  # should knock out maxTime
-        return d
+        yield d
+
+        self.assertTrue(
+            {'stdout': nl('hello\n')} not in b.updates, b.show())
+        self.assertTrue({'rc': FATAL_RC} in b.updates, b.show())
 
     @compat.skipUnlessPlatformIs("posix")
+    @defer.inlineCallbacks
     def test_stdin_closed(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(b,
@@ -330,67 +344,61 @@ class TestRunProcess(BasedirMixin, unittest.TestCase):
                                   # if usePTY=True, stdin is never closed
                                   usePTY=False,
                                   logEnviron=False)
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            self.assertTrue({'rc': 0} in b.updates, b.show())
-        d.addCallback(check)
-        return d
+        self.assertTrue({'rc': 0} in b.updates, b.show())
 
-    @compat.usesFlushLoggedErrors
+    @defer.inlineCallbacks
     def test_startCommand_exception(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(b, ['whatever'], self.basedir)
 
         # set up to cause an exception in _startCommand
         def _startCommand(*args, **kwargs):
-            raise RuntimeError()
+            raise RuntimeError('error message')
         s._startCommand = _startCommand
 
-        d = s.start()
+        try:
+            yield s.start()
+        except AbandonChain:
+            pass
 
-        def check(err):
-            err.trap(AbandonChain)
-            stderr = []
-            # Here we're checking that the exception starting up the command
-            # actually gets propagated back to the master in stderr.
-            for u in b.updates:
-                if 'stderr' in u:
-                    stderr.append(u['stderr'])
-            stderr = "".join(stderr)
-            self.assertTrue("RuntimeError" in stderr, stderr)
-        d.addBoth(check)
-        d.addBoth(lambda _: self.flushLoggedErrors())
-        return d
+        stderr = []
+        # Here we're checking that the exception starting up the command
+        # actually gets propagated back to the master in stderr.
+        for u in b.updates:
+            if 'stderr' in u:
+                stderr.append(u['stderr'])
+        stderr = ''.join(stderr)
+        self.assertTrue(stderr.startswith('error in RunProcess._startCommand (error message)'))
 
+        yield self.flushLoggedErrors()
+
+    @defer.inlineCallbacks
     def testLogEnviron(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(b, stdoutCommand('hello'), self.basedir,
                                   environ={"FOO": "BAR"})
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            headers = "".join([list(update.values())[0]
-                               for update in b.updates if list(update) == ["header"]])
-            self.assertTrue("FOO=BAR" in headers, "got:\n" + headers)
-        d.addCallback(check)
-        return d
+        headers = "".join([list(update.values())[0]
+                           for update in b.updates if list(update) == ["header"]])
+        self.assertTrue("FOO=BAR" in headers, "got:\n" + headers)
 
+    @defer.inlineCallbacks
     def testNoLogEnviron(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(b, stdoutCommand('hello'), self.basedir,
                                   environ={"FOO": "BAR"}, logEnviron=False)
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            headers = "".join([list(update.values())[0]
-                               for update in b.updates if list(update) == ["header"]])
-            self.assertTrue("FOO=BAR" not in headers, "got:\n" + headers)
-        d.addCallback(check)
-        return d
+        headers = "".join([list(update.values())[0]
+                           for update in b.updates if list(update) == ["header"]])
+        self.assertTrue("FOO=BAR" not in headers, "got:\n" + headers)
 
+    @defer.inlineCallbacks
     def testEnvironExpandVar(self):
         b = FakeWorkerForBuilder(self.basedir)
         environ = {"EXPND": "-${PATH}-",
@@ -399,68 +407,91 @@ class TestRunProcess(BasedirMixin, unittest.TestCase):
         s = runprocess.RunProcess(
             b, stdoutCommand('hello'), self.basedir, environ=environ)
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            headers = "".join([list(update.values())[0]
-                               for update in b.updates if list(update) == ["header"]])
-            self.assertTrue("EXPND=-$" not in headers, "got:\n" + headers)
-            self.assertTrue("DOESNT_FIND=--" in headers, "got:\n" + headers)
-            self.assertTrue(
-                "DOESNT_EXPAND=-${---}-" in headers, "got:\n" + headers)
-        d.addCallback(check)
-        return d
+        headers = "".join([list(update.values())[0]
+                           for update in b.updates if list(update) == ["header"]])
+        self.assertTrue("EXPND=-$" not in headers, "got:\n" + headers)
+        self.assertTrue("DOESNT_FIND=--" in headers, "got:\n" + headers)
+        self.assertTrue(
+            "DOESNT_EXPAND=-${---}-" in headers, "got:\n" + headers)
 
+    @defer.inlineCallbacks
     def testUnsetEnvironVar(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(b, stdoutCommand('hello'), self.basedir,
                                   environ={"PATH": None})
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            headers = "".join([list(update.values())[0]
-                               for update in b.updates if list(update) == ["header"]])
-            self.assertFalse(
-                re.match('\bPATH=', headers), "got:\n" + headers)
-        d.addCallback(check)
-        return d
+        headers = "".join([list(update.values())[0]
+                           for update in b.updates if list(update) == ["header"]])
+        self.assertFalse(
+            re.match('\bPATH=', headers), "got:\n" + headers)
 
+    @defer.inlineCallbacks
     def testEnvironPythonPath(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(b, stdoutCommand('hello'), self.basedir,
                                   environ={"PYTHONPATH": 'a'})
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            headers = "".join([list(update.values())[0]
-                               for update in b.updates if list(update) == ["header"]])
-            self.assertFalse(re.match('\bPYTHONPATH=a%s' % (os.pathsep), headers),
-                             "got:\n" + headers)
-        d.addCallback(check)
-        return d
+        headers = "".join([list(update.values())[0]
+                           for update in b.updates if list(update) == ["header"]])
+        self.assertFalse(re.match('\bPYTHONPATH=a{0}'.format(os.pathsep), headers),
+                         "got:\n" + headers)
 
+    @defer.inlineCallbacks
     def testEnvironArray(self):
         b = FakeWorkerForBuilder(self.basedir)
         s = runprocess.RunProcess(b, stdoutCommand('hello'), self.basedir,
                                   environ={"FOO": ['a', 'b']})
 
-        d = s.start()
+        yield s.start()
 
-        def check(ign):
-            headers = "".join([list(update.values())[0]
-                               for update in b.updates if list(update) == ["header"]])
-            self.assertFalse(re.match('\bFOO=a{0}b\b'.format(os.pathsep), headers),
-                             "got:\n" + headers)
-        d.addCallback(check)
-        return d
+        headers = "".join([list(update.values())[0]
+                           for update in b.updates if list(update) == ["header"]])
+        self.assertFalse(re.match('\bFOO=a{0}b\b'.format(os.pathsep), headers),
+                         "got:\n" + headers)
 
     def testEnvironInt(self):
         b = FakeWorkerForBuilder(self.basedir)
-        self.assertRaises(RuntimeError, lambda:
-                          runprocess.RunProcess(b, stdoutCommand('hello'), self.basedir,
-                                                environ={"BUILD_NUMBER": 13}))
+        with self.assertRaises(RuntimeError):
+            runprocess.RunProcess(b, stdoutCommand('hello'), self.basedir,
+                                  environ={"BUILD_NUMBER": 13})
+
+    def _test_spawnAsBatch(self, cmd, comspec):
+
+        def spawnProcess(processProtocol, executable, args=(), env=None,
+                         path=None, uid=None, gid=None, usePTY=False, childFDs=None):
+            self.assertTrue(args[0].lower().endswith("cmd.exe"),
+                            "{0} is not cmd.exe".format(args[0]))
+
+        self.patch(runprocess.reactor, "spawnProcess", spawnProcess)
+        tempEnviron = os.environ.copy()
+        if 'COMSPEC' not in tempEnviron:
+            tempEnviron['COMSPEC'] = comspec
+        self.patch(os, "environ", tempEnviron)
+        b = FakeWorkerForBuilder(self.basedir)
+        s = runprocess.RunProcess(b, cmd, self.basedir)
+        s.pp = runprocess.RunProcessPP(s)
+        s.deferred = defer.Deferred()
+        d = s._spawnAsBatch(s.pp, s.command, "args",
+                            tempEnviron, "path", False)
+        return d
+
+    def test_spawnAsBatchCommandString(self):
+        return self._test_spawnAsBatch("dir c:/", "cmd.exe")
+
+    def test_spawnAsBatchCommandList(self):
+        return self._test_spawnAsBatch(stdoutCommand('hello'), "cmd.exe /c")
+
+    def test_spawnAsBatchCommandWithNonAscii(self):
+        return self._test_spawnAsBatch(u"echo \u6211", "cmd.exe")
+
+    def test_spawnAsBatchCommandListWithNonAscii(self):
+        return self._test_spawnAsBatch(['echo', u"\u6211"], "cmd.exe /c")
 
 
 class TestPOSIXKilling(BasedirMixin, unittest.TestCase):
@@ -495,7 +526,7 @@ class TestPOSIXKilling(BasedirMixin, unittest.TestCase):
         self.tearDownBasedir()
 
     def newPidfile(self):
-        pidfile = os.path.abspath("test-%d.pid" % len(self.pidfiles))
+        pidfile = os.path.abspath("test-{0}.pid".format(len(self.pidfiles)))
         if os.path.exists(pidfile):
             os.unlink(pidfile)
         self.pidfiles.append(pidfile)
@@ -508,11 +539,13 @@ class TestPOSIXKilling(BasedirMixin, unittest.TestCase):
 
         def poll():
             if reactor.seconds() > until:
-                d.errback(RuntimeError("pidfile %s never appeared" % pidfile))
+                d.errback(RuntimeError(
+                    "pidfile {0} never appeared".format(pidfile)))
                 return
             if os.path.exists(pidfile):
                 try:
-                    pid = int(open(pidfile).read())
+                    with open(pidfile) as f:
+                        pid = int(f.read())
                 except (IOError, TypeError, ValueError):
                     pid = None
 
@@ -527,10 +560,10 @@ class TestPOSIXKilling(BasedirMixin, unittest.TestCase):
         try:
             os.kill(pid, 0)
         except OSError:
-            self.fail("pid %d still alive" % (pid,))
+            self.fail("pid {0} still alive".format(pid))
 
     def assertDead(self, pid, timeout=5):
-        log.msg("checking pid %r" % (pid,))
+        log.msg("checking pid {0!r}".format(pid))
 
         def check():
             try:
@@ -551,7 +584,7 @@ class TestPOSIXKilling(BasedirMixin, unittest.TestCase):
             time.sleep(0.01)
             if check():
                 return
-        self.fail("pid %d still alive after %ds" % (pid, timeout))
+        self.fail("pid {0} still alive after {1}s".format(pid, timeout))
 
     # tests
 
@@ -642,6 +675,7 @@ class TestPOSIXKilling(BasedirMixin, unittest.TestCase):
         return self.do_test_pgroup(usePTY=False, useProcGroup=False,
                                    expectChildSurvival=True)
 
+    @defer.inlineCallbacks
     def do_test_pgroup(self, usePTY, useProcGroup=True,
                        expectChildSurvival=False):
         # test that a process group gets killed
@@ -673,16 +707,13 @@ class TestPOSIXKilling(BasedirMixin, unittest.TestCase):
         pidfiles_d.addCallback(kill)
 
         # check that both processes are dead after RunProcess is done
-        d = defer.gatherResults([pidfiles_d, runproc_d])
+        yield defer.gatherResults([pidfiles_d, runproc_d])
 
-        def check_dead(_):
-            self.assertDead(self.parent_pid)
-            if expectChildSurvival:
-                self.assertAlive(self.child_pid)
-            else:
-                self.assertDead(self.child_pid)
-        d.addCallback(check_dead)
-        return d
+        self.assertDead(self.parent_pid)
+        if expectChildSurvival:
+            self.assertAlive(self.child_pid)
+        else:
+            self.assertDead(self.child_pid)
 
     def test_double_fork_usePTY(self):
         return self.do_test_double_fork(usePTY=True)
@@ -696,6 +727,7 @@ class TestPOSIXKilling(BasedirMixin, unittest.TestCase):
         return self.do_test_double_fork(usePTY=False, useProcGroup=False,
                                         expectChildSurvival=True)
 
+    @defer.inlineCallbacks
     def do_test_double_fork(self, usePTY, useProcGroup=True,
                             expectChildSurvival=False):
         # when a spawned process spawns another process, and then dies itself
@@ -729,17 +761,200 @@ class TestPOSIXKilling(BasedirMixin, unittest.TestCase):
         pidfiles_d.addCallback(kill)
 
         # check that both processes are dead after RunProcess is done
-        d = defer.gatherResults([pidfiles_d, runproc_d])
+        yield defer.gatherResults([pidfiles_d, runproc_d])
 
-        def check_dead(_):
-            self.assertDead(self.parent_pid)
-            if expectChildSurvival:
-                self.assertAlive(self.child_pid)
-            else:
-                self.assertDead(self.child_pid)
-        d.addCallback(check_dead)
+        self.assertDead(self.parent_pid)
+        if expectChildSurvival:
+            self.assertAlive(self.child_pid)
+        else:
+            self.assertDead(self.child_pid)
+
+
+class TestWindowsKilling(BasedirMixin, unittest.TestCase):
+
+    if runtime.platformType != "win32":
+        skip = "not a Windows platform"
+
+    def setUp(self):
+        self.pidfiles = []
+        self.setUpBasedir()
+
+    def tearDown(self):
+        # make sure all of the subprocesses are dead
+        for pidfile in self.pidfiles:
+            if not os.path.exists(pidfile):
+                continue
+            with open(pidfile) as f:
+                pid = f.read()
+            if not pid:
+                return
+            pid = int(pid)
+            try:
+                psutil.Process(pid).kill()
+            except psutil.NoSuchProcess:
+                pass
+            while psutil.pid_exists(pid):
+                time.sleep(0.01)
+
+        # and clean up leftover pidfiles
+        for pidfile in self.pidfiles:
+            if os.path.exists(pidfile):
+                os.unlink(pidfile)
+
+        self.tearDownBasedir()
+
+    def new_pid_file(self):
+        pidfile = os.path.abspath("test-{0}.pid".format(len(self.pidfiles)))
+        if os.path.exists(pidfile):
+            os.unlink(pidfile)
+        self.pidfiles.append(pidfile)
+        return pidfile
+
+    def wait_for_pidfile(self, pidfile):
+        # wait for a pidfile, and return the pid via a Deferred
+        until = time.time() + 10
+        d = defer.Deferred()
+
+        def poll():
+            if reactor.seconds() > until:
+                d.errback(RuntimeError(
+                    "pidfile {0} never appeared".format(pidfile)))
+                return
+            if os.path.exists(pidfile):
+                try:
+                    with open(pidfile) as f:
+                        pid = int(f.read())
+                except (IOError, TypeError, ValueError):
+                    pid = None
+
+                if pid is not None:
+                    d.callback(pid)
+                    return
+            reactor.callLater(0.01, poll)
+        poll()  # poll right away
         return d
 
+    def assert_alive(self, pid):
+        if not psutil.pid_exists(pid):
+            self.fail("pid {0} dead, but expected it to be alive".format(pid))
+
+    def assert_dead(self, pid, timeout=5):
+        log.msg("checking pid {0!r}".format(pid))
+
+        # check immediately
+        if not psutil.pid_exists(pid):
+            return
+
+        # poll every 100'th of a second; this allows us to test for
+        # processes that have been killed, but where the signal hasn't
+        # been delivered yet
+        until = time.time() + timeout
+        while time.time() < until:
+            time.sleep(0.01)
+            if not psutil.pid_exists(pid):
+                return
+        self.fail("pid {0} still alive after {1}s".format(pid, timeout))
+
+    # tests
+
+    @defer.inlineCallbacks
+    def test_simple(self, interrupt_signal=None):
+
+        # test a simple process that just sleeps waiting to die
+        pidfile = self.new_pid_file()
+
+        b = FakeWorkerForBuilder(self.basedir)
+        s = runprocess.RunProcess(b, scriptCommand('write_pidfile_and_sleep', pidfile),
+                                  self.basedir)
+        if interrupt_signal is not None:
+            s.interruptSignal = interrupt_signal
+        runproc_d = s.start()
+
+        pid = yield self.wait_for_pidfile(pidfile)
+
+        self.assert_alive(pid)
+
+        # test that the process is still alive and tell the RunProcess object to kill it
+        s.kill("diaf")
+
+        yield runproc_d
+        self.assert_dead(pid)
+
+    @defer.inlineCallbacks
+    def test_sigterm(self):
+        # Tests that the process will receive SIGTERM if sigtermTimeout is not None
+        pidfile = self.new_pid_file()
+
+        b = FakeWorkerForBuilder(self.basedir)
+        s = runprocess.RunProcess(b, scriptCommand('write_pidfile_and_sleep', pidfile),
+                                  self.basedir, sigtermTime=1)
+
+        taskkill_calls = []
+        orig_taskkill = s._taskkill
+
+        def mock_taskkill(pid, force):
+            taskkill_calls.append(force)
+            orig_taskkill(pid, force)
+        s._taskkill = mock_taskkill
+
+        runproc_d = s.start()
+        pid = yield self.wait_for_pidfile(pidfile)
+        # test that the process is still alive
+        self.assert_alive(pid)
+        # and tell the RunProcess object to kill it
+        s.kill("diaf")
+
+        yield runproc_d
+        self.assertEqual(taskkill_calls, [False, True])
+        self.assert_dead(pid)
+
+    @defer.inlineCallbacks
+    def test_with_child(self):
+        # test that a process group gets killed
+        parent_pidfile = self.new_pid_file()
+        child_pidfile = self.new_pid_file()
+
+        b = FakeWorkerForBuilder(self.basedir)
+        s = runprocess.RunProcess(b, scriptCommand('spawn_child', parent_pidfile, child_pidfile),
+                                  self.basedir)
+        runproc_d = s.start()
+
+        # wait for both processes to start up, then call s.kill
+        parent_pid = yield self.wait_for_pidfile(parent_pidfile)
+        child_pid = yield self.wait_for_pidfile(child_pidfile)
+
+        s.kill("diaf")
+
+        yield runproc_d
+
+        self.assert_dead(parent_pid)
+        self.assert_dead(child_pid)
+
+    @defer.inlineCallbacks
+    def test_with_child_parent_dies(self):
+        # when a spawned process spawns another process, and then dies itself
+        # (either intentionally or accidentally), we can't kill the child process.
+        # In the future we should be able to fix this as Windows has CREATE_NEW_PROCESS_GROUP.
+        parent_pidfile = self.new_pid_file()
+        child_pidfile = self.new_pid_file()
+
+        b = FakeWorkerForBuilder(self.basedir)
+        s = runprocess.RunProcess(b, scriptCommand('double_fork', parent_pidfile, child_pidfile),
+                                  self.basedir)
+        runproc_d = s.start()
+
+        # wait for both processes to start up, then call s.kill
+        parent_pid = yield self.wait_for_pidfile(parent_pidfile)
+        child_pid = yield self.wait_for_pidfile(child_pidfile)
+
+        s.kill("diaf")
+
+        # check that both processes are dead after RunProcess is done
+        yield runproc_d
+
+        self.assert_dead(parent_pid)
+        self.assert_alive(child_pid)
+
 
 class TestLogging(BasedirMixin, unittest.TestCase):
 
@@ -791,6 +1006,15 @@ class TestLogging(BasedirMixin, unittest.TestCase):
         s._addToBuffers('stdout', data)
         self.assertEqual(len(b.updates), 1)
 
+    def testSendLog(self):
+        b = FakeWorkerForBuilder(self.basedir)
+        s = runprocess.RunProcess(b, stdoutCommand('hello'), self.basedir)
+        s._addToBuffers(('log', 'stdout'), 'hello ')
+        s._sendBuffers()
+        self.assertEqual(b.updates, [
+            {'log': ('stdout', 'hello ')},
+        ])
+
 
 class TestLogFileWatcher(BasedirMixin, unittest.TestCase):
 
@@ -807,16 +1031,46 @@ class TestLogFileWatcher(BasedirMixin, unittest.TestCase):
 
     def test_statFile_missing(self):
         rp = self.makeRP()
-        if os.path.exists('statfile.log'):
-            os.remove('statfile.log')
-        lf = runprocess.LogFileWatcher(rp, 'test', 'statfile.log', False)
-        self.assertFalse(lf.statFile(), "statfile.log doesn't exist")
+        test_filename = 'test_runprocess_test_statFile_missing.log'
+        if os.path.exists(test_filename):
+            os.remove(test_filename)
+        lf = runprocess.LogFileWatcher(rp, 'test', test_filename, False)
+        self.assertFalse(lf.statFile(), "{} doesn't exist".format(test_filename))
 
     def test_statFile_exists(self):
         rp = self.makeRP()
-        open('statfile.log', 'w').write('hi')
-        lf = runprocess.LogFileWatcher(rp, 'test', 'statfile.log', False)
-        st = lf.statFile()
-        self.assertEqual(
-            st and st[2], 2, "statfile.log exists and size is correct")
-        os.remove('statfile.log')
+        test_filename = 'test_runprocess_test_statFile_exists.log'
+        try:
+            with open(test_filename, 'w') as f:
+                f.write('hi')
+            lf = runprocess.LogFileWatcher(rp, 'test', test_filename, False)
+            st = lf.statFile()
+            self.assertEqual(
+                st and st[2], 2, "statfile.log exists and size is correct")
+        finally:
+            os.remove(test_filename)
+
+    def test_invalid_utf8(self):
+        # create the log file watcher first
+        rp = self.makeRP()
+        test_filename = 'test_runprocess_test_invalid_utf8.log'
+
+        try:
+            lf = runprocess.LogFileWatcher(rp, 'test', test_filename,
+                                           follow=False, poll=False)
+            # now write to the log file
+            INVALID_UTF8 = b'before\xffafter'
+            with open(test_filename, 'wb') as f:
+                f.write(INVALID_UTF8)
+            # the watcher picks up the changed log
+            lf.poll()
+            # flush she buffer
+            rp._sendBuffers()
+            # the log file content was captured and the invalid byte replaced with \ufffd (the
+            # replacement character, often a black diamond with a white question mark)
+            REPLACED = u'before\ufffdafter'
+            self.assertEqual(rp.builder.updates, [{'log': ('test', REPLACED)}])
+
+        finally:
+            lf.stop()
+            os.remove(f.name)
